﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Reflection;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Data;
using System.Windows.Media;
using Styx.Helpers;

namespace NuokRealmHop
{
    //Full credit to Apoc
    internal class ConfigWindow : Window
    {
        private readonly Grid _mainCanvas = new Grid();

        /// <summary>
        /// </summary>
        /// <param name="title">Window Title</param>
        /// <param name="logo">Logo</param>
        /// <param name="logoTagLine"></param>
        /// <param name="width">Frame Width</param>
        /// <param name="height">Frame Height</param>
        public ConfigWindow(string title, string logo, string logoTagLine, int width, int height)
        {
            WindowStyle = WindowStyle.SingleBorderWindow;
            Topmost = true;
            Width = width;
            Height = height;
            Title = title;


            // Store _mainCanvas outside this so we don't have to cast and whatnot.
            Content = _mainCanvas;
            _mainCanvas.RowDefinitions.Add(new RowDefinition());
            _mainCanvas.RowDefinitions.Add(new RowDefinition {Height = new GridLength(50)});

            Background = Brushes.LightGray;

            // Add tab controls...
            var tabs = new TabControl();
            Grid.SetRow(tabs, 0);
            _mainCanvas.Children.Add(tabs);

            var mainSettingsSource = NSettings.Instance;

            #region Tabs

            var raiDifficulty = GenerateTab("Main Window", mainSettingsSource);
            tabs.Items.Add(raiDifficulty);

            #endregion

            #region "Logo"

            var logoPanel = new StackPanel();
            Grid.SetRow(logoPanel, 1);
            _mainCanvas.Children.Add(logoPanel);

            var logoMain = new TextBlock
            {
                Text = logo,
                Foreground = new SolidColorBrush(Colors.White),
                FontSize = 20,
                FontFamily = new FontFamily("Impact"),
                Padding = new Thickness(5, 0, 0, 0)
            };

            logoPanel.Children.Add(logoMain);

            var logoTag = new TextBox
            {
                Text = logoTagLine,
                Foreground = new SolidColorBrush(Colors.White),
                FontWeight = FontWeights.Bold,
                Padding = new Thickness(5),
                BorderThickness = new Thickness(0)
            };

            logoPanel.Children.Add(logoTag);

            #endregion
        }

        private TabItem GenerateTab(string tabtitle, object source)
        {
            var tab = new TabItem {Header = tabtitle};
            var tabPanel = new StackPanel();
            var sv = new ScrollViewer
            {
                Content = tabPanel,
                BorderBrush = Brushes.LightGray,
                BorderThickness = new Thickness(2)
            };
            tab.Content = sv;

            var settingsProperties =
                source.GetType()
                    .GetProperties()
                    .Where(p => p.GetCustomAttributes(false).Any(a => a is SettingAttribute));

            BuildSettings(tabPanel, settingsProperties, source);

            return tab;
        }

        private static T GetAttribute<T>(PropertyInfo pi) where T : Attribute
        {
            var attr = pi.GetCustomAttributes(false).FirstOrDefault(a => a is T);
            return attr as T;
        }

        private void BuildSettings(Panel tabPanel, IEnumerable<PropertyInfo> settings, object source)
        {
            var categories = new Dictionary<string, StackPanel>();

            foreach (var pi in settings.OrderBy(p => p.PropertyType.Name))
            {
                // First get some display stuff.
                // By default, any settings w/o a category set, go into "Misc"
                var category = GetAttribute<CategoryAttribute>(pi) != null
                    ? GetAttribute<CategoryAttribute>(pi).Category
                    : "Miscellaneous";
                var displayName = GetAttribute<DisplayNameAttribute>(pi) != null
                    ? GetAttribute<DisplayNameAttribute>(pi).DisplayName
                    // Default to the property name if no display name is given.
                    : pi.Name;
                var description = GetAttribute<DescriptionAttribute>(pi) != null
                    ? GetAttribute<DescriptionAttribute>(pi).Description
                    : null;

                if (!categories.ContainsKey(category))
                {
                    categories.Add(category, new StackPanel());
                }

                var group = categories[category];

                //Logger.Write(displayName + " -> " + description);

                var returnType = pi.PropertyType;

                // Deal with enums in a "special" way. The typecode for any enum will be int32 by default (unless marked as something else)
                // So we really want a dropdown, not a textbox.
                if (returnType.IsEnum)
                {
                    AddComboBoxForEnum(group, source, pi, displayName, description, returnType);
                    continue;
                }

                // Easiest way to blanket-statement a bunch of editable values. (Quite a few will just be textbox editors)
                switch (Type.GetTypeCode(returnType))
                {
                    case TypeCode.Boolean:
                        AddCheckbox(group, source, pi, displayName, description);
                        break;
                    case TypeCode.Char:
                    case TypeCode.SByte:
                    case TypeCode.Byte:
                    case TypeCode.Int16:
                    case TypeCode.UInt16:
                    case TypeCode.Int32:
                    case TypeCode.UInt32:
                    case TypeCode.Int64:
                    case TypeCode.UInt64:
                    case TypeCode.Single:
                    case TypeCode.Double:
                    case TypeCode.Decimal:
                        AddSlider(group, source, pi, displayName, description);
                        break;
                    case TypeCode.String:
                        AddEditBox(group, source, pi, displayName, description);
                        break;
                }
            }

            foreach (var sp in categories.OrderBy(kv => kv.Key))
            {
                var gb = new Expander
                {
                    Content = sp.Value,
                    Header = sp.Key,
                    ExpandDirection = ExpandDirection.Down,
                    IsExpanded = true
                };
                tabPanel.Children.Add(gb);
            }
        }

        private void AddBinding(FrameworkElement ctrl, string xpath, object source, PropertyInfo bindTo,
            DependencyProperty depProp)
        {
            var b = new Binding(xpath)
            {
                Source = source,
                Path = new PropertyPath(bindTo.Name),
                Mode = BindingMode.TwoWay
            };
            ctrl.SetBinding(depProp, b);
        }

        private void AddCheckbox(Panel ctrl, object source, PropertyInfo bindTo, string label, string tooltip)
        {
            var display = new StackPanel {Orientation = Orientation.Horizontal, Width = ctrl.Width};
            var l = new Label {Content = label, Margin = new Thickness(5, 5, 5, 3)};
            var cb = new CheckBox
            {
                ToolTip = !string.IsNullOrEmpty(tooltip) ? tooltip : null,
                Margin = new Thickness(5, 5, 5, 3)
            };

            // And the binding so we don't have to do a lot of nasty event handling.
            AddBinding(cb, "IsChecked", source, bindTo, ToggleButton.IsCheckedProperty);

            display.Children.Add(l);
            display.Children.Add(cb);

            ctrl.Children.Add(display);
        }

        private void AddSlider(Panel ctrl, object source, PropertyInfo bindTo, string label, string tooltip)
        {
            var display = new StackPanel {Orientation = Orientation.Horizontal, Width = ctrl.Width};
            var l = new Label {Content = label, ToolTip = tooltip, Margin = new Thickness(5, 5, 5, 3)};
            var sldLbl = new Label {Margin = new Thickness(5, 5, 5, 3)};

            var s = new Slider();
            // Find min/max
            var attr = GetAttribute<LimitAttribute>(bindTo);
            if (attr != null)
            {
                s.Maximum = attr.High;
                s.Minimum = attr.Low;
                s.TickFrequency = Math.Abs(attr.High - attr.Low)/100;
                s.SmallChange = s.TickFrequency;
                s.LargeChange = s.TickFrequency*10f;
            }
            s.MinWidth = 150;
            s.TickPlacement = TickPlacement.BottomRight;
            s.ToolTip = tooltip;
            AddBinding(s, "Value", source, bindTo, RangeBase.ValueProperty);

            var b = new Binding("Value") {Source = s, Path = new PropertyPath("Value"), Mode = BindingMode.TwoWay};
            sldLbl.SetBinding(ContentProperty, b);
            sldLbl.ContentStringFormat = "N2";
            sldLbl.Width = 38;

            display.Children.Add(l);
            display.Children.Add(sldLbl);
            display.Children.Add(s);

            ctrl.Children.Add(display);
        }

        private void AddEditBox(Panel ctrl, object source, PropertyInfo bindTo, string label, string tooltip)
        {
            // This is a bit tricky. We want to stack the edit box to the right of the label.
            // We do this via another stack panel, with a changed "stack" way.

            var display = new StackPanel {Orientation = Orientation.Horizontal, Width = ctrl.Width};
            var l = new Label {Content = label, Margin = new Thickness(5, 5, 5, 3)};

            var tb = new TextBox
            {
                ToolTip = tooltip,
                Background = Background,
                BorderBrush = new SolidColorBrush(Colors.LightGray),
                Margin = new Thickness(2, 3, 5, 3),
                MinWidth = 50
            };

            // Add the textbox/label to the stack panel so we can have it side-to-side.
            display.Children.Add(tb);
            display.Children.Add(l);

            // Don't forget the damned binding.
            AddBinding(tb, "Text", source, bindTo, TextBox.TextProperty);

            // And add it to the main control.
            ctrl.Children.Add(display);
        }

        private void AddComboBoxForEnum(Panel ctrl, object source, PropertyInfo bindTo, string label, string tooltip,
            Type enumType)
        {
            var display = new StackPanel {Orientation = Orientation.Horizontal, Width = ctrl.Width};
            var l = new Label {Content = label, Margin = new Thickness(5, 5, 5, 3)};

            var cb = new ComboBox
            {
                ToolTip = tooltip,
                Background = Background,
                BorderBrush = new SolidColorBrush(Colors.LightGray),
                BorderThickness = new Thickness(2)
            };
            foreach (var val in Enum.GetValues(enumType))
            {
                cb.Items.Add(val);
            }

            AddBinding(cb, "SelectedItem", source, bindTo, Selector.SelectedItemProperty);

            display.Children.Add(l);
            display.Children.Add(cb);

            ctrl.Children.Add(display);
        }
    }
}