using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using System.Windows.Media;
using CommonBehaviors.Actions;
using Styx;
using Styx.Common;
using Styx.CommonBot.Coroutines;
using Styx.WoWInternals;

namespace NuokRealmHop
{
    public class LfgList
    {
        private static IEnumerable<int> GetAvailableCategorieIds
        {
            get
            {
                using (new PerformanceLogger(NSettings.Instance.DebugPerformace, "GetAvailableCategorieIds"))
                {
                    using (StyxWoW.Memory.AcquireFrame())
                    {
                        const string getIdsLua = @"local categories = C_LFGList.GetAvailableCategories();
local results = {}
for i = 1, #categories do        
local id = categories[i]
if (id ~= nil) then 
table.insert(results, id) 
end 
end 
return unpack(results)";
                        var l = new List<int>();
                        var getResultIds = Lua.GetReturnValues(getIdsLua);
                        l.AddRange(getResultIds.Select(Lua.ParseLuaValue<int>));
                        return l;
                    }
                }
            }
        }

        public static List<CategoryInfo> GetAvailableCategories
        {
            get
            {
                using (new PerformanceLogger(NSettings.Instance.DebugPerformace, "GetAvailableCategories"))
                {
                    using (StyxWoW.Memory.AcquireFrame())
                    {
                        var x = GetAvailableCategorieIds.Select(GetCategoryInfo).ToList();
                        if (!NSettings.Instance.DebugDetailed) return x;
                        foreach (var r in x)
                        {
                            Log($"GetAvailableCategories: {r}");
                        }
                        return x;
                    }
                }
            }
        }


        /// <summary>
        ///     Retrieve the search information for the last search only
        /// </summary>
        public static List<SearchResultInfo> GetSearchResults
        {
            get
            {
                using (new PerformanceLogger(NSettings.Instance.DebugPerformace, "GetSearchResults"))
                {
                    const string getIdsLua = @"local numResults, resultIDTable = C_LFGList.GetSearchResults();
local results = {}
for i = 1, numResults do        
local id = resultIDTable[i]
if (id ~= nil) then 
table.insert(results, id) 
end
end
return unpack(results)";

                    var l = new List<int>();
                    using (StyxWoW.Memory.AcquireFrame())
                    {
                        var getResultIds = Lua.GetReturnValues(getIdsLua);

                        l.AddRange(getResultIds.Select(Lua.ParseLuaValue<int>));
                    }
                    var x = (from i in l where i != 0 select GetSearchResultInfo(i)).ToList();
                    if (!NSettings.Instance.DebugDetailed) return x;
                    foreach (var r in x)
                    {
                        Log($"GetSearchResults: {r}");
                    }
                    return x;
                }
            }
        }

        public static int NumActiveAppliications => Lua.GetReturnVal<int>("return C_LFGList.GetNumApplications()", 1);

        /// <summary>
        ///     Gets all the currently applied for groups
        /// </summary>
        public static List<ApplicationInfo> GetApplicationInfos
        {
            get
            {
                using (new PerformanceLogger(NSettings.Instance.DebugPerformace, "GetApplicationInfos"))
                {
                    var l = new List<ApplicationInfo>();

                    using (StyxWoW.Memory.AcquireFrame())
                    {
                        var apps = GetApplicationIds;
                        l.AddRange(apps.Select(GetApplicationInfo));
                    }

                    if (!NSettings.Instance.DebugDetailed) return l;
                    foreach (var r in l)
                    {
                        Log($"GetApplicationInfos: {r}");
                    }

                    return l;
                }
            }
        }

        /// <summary>
        ///     Gets the players current realm name
        /// </summary>
        public static string GetRealmName
        {
            get
            {
                using (new PerformanceLogger(NSettings.Instance.DebugPerformace, "GetRealmName"))
                {
                    string realm;
                    using (StyxWoW.Memory.AcquireFrame())
                    {
                        string lua;
                        var gi = StyxWoW.Me.GroupInfo;
                        var meLeader = gi.GroupLeader == StyxWoW.Me;
                        if (gi.IsInRaid && !meLeader)
                            lua = @"local index = GetNumGroupMembers();
for i = 1,index do
if UnitIsGroupLeader('raid' ..i) then
local name, realm = UnitName('raid' ..i);
if realm == '' or realm == nil then
realm = GetRealmName();
end
return realm;
end
end";
                        else if (gi.IsInParty && !meLeader)
                            lua =
                                @"for i = 1,4 do
if UnitIsGroupLeader('party' ..i) then
local name, realm = UnitName('party' ..i);
if realm == '' or realm == nil then
realm = GetRealmName();
end
return realm;
end 
end";
                        else lua = "return GetRealmName();";
                        realm = Lua.GetReturnVal<string>(lua, 0);
                    }

                    if (NSettings.Instance.DebugDetailed && realm != null)
                        Log($"GetRealmName: {realm}");

                    return string.IsNullOrEmpty(realm) ? Lua.GetReturnVal<string>("return GetRealmName();", 0) : realm;
                }
            }
        }

        private static IEnumerable<int> GetApplicationIds
        {
            get
            {
                var l = new List<int>();
                const string lua = @"local apps = C_LFGList.GetApplications(); 
local results = {}
for i = 1, #apps do        
local id = apps[i]
if (id ~= nil) then 
table.insert(results, id) 
end
end
return unpack(results)";
                var getResultIds = Lua.GetReturnValues(lua);
                l.AddRange(getResultIds.Select(Lua.ParseLuaValue<int>));
                return l;
            }
        }

        /// <summary>
        ///     Get all avalivble activity ids
        /// </summary>
        private static List<int> GetAvailableActivitiesIds
        {
            get
            {
                // /dump C_LFGList.GetAvailableActivities()
                const string lua = "return unpack(C_LFGList.GetAvailableActivities());";
                var r = Lua.GetReturnValues(lua);
                return r.Select(Lua.ParseLuaValue<int>).ToList();
            }
        }

        /// <summary>
        ///     Get all avalible activities
        /// </summary>
        public static List<ActivityInfo> GetAvailableActivities
        {
            get
            {
                using (new PerformanceLogger(NSettings.Instance.DebugDetailed, "GetAvailableActivities"))
                {
                    var r = new List<ActivityInfo>();
                    using (StyxWoW.Memory.AcquireFrame())
                    {
                        var ids = GetAvailableActivitiesIds;
                        r.AddRange(ids.Select(GetActivityInfo));
                    }
                    if (!NSettings.Instance.DebugDetailed)
                        return r;

                    Log("GetAvailableActivities");
                    foreach (var a in r)
                    {
                        Log($"  {a}");
                    }
                    return r;
                }
            }
        }

        /// <summary>
        ///     Opens the LFG Frame in game
        /// </summary>
        public static void LFGLiftFrame_Show()
        {
            Lua.DoString("PVEFrame_ShowFrame('GroupFinderFrame', LFGListPVEStub)");
        }

        private static void Log(string format, params object[] args)
        {
            Logging.WriteDiagnostic(Colors.Goldenrod, "LFGList: " + format, args);
        }

        /// <summary>
        ///     Applies to join the group
        /// </summary>
        /// <param name="id">Group Id</param>
        public static void ApplyToGroup(int id)
        {
            Lua.DoString($"local tank, heal, dd = C_LFGList.GetAvailableRoles(); C_LFGList.ApplyToGroup({id}, '', tank, heal, dd);");
        }

        /// <summary>
        ///     Applies to join the group
        /// </summary>
        /// <param name="application"></param>
        public static void ApplyToGroup(SearchResultInfo application)
        {
            Log($"Applying to Group {application} on {application.Realm} (AutoInvite: {application.AutoInvite})");
            ApplyToGroup(application.Id);
        }

        /// <summary>
        ///     Accepts group invite pop ups
        /// </summary>
        public static void AcceptPopUps()
        {
            if (NSettings.Instance.DebugDetailed)
                Log("Accepting PopUps");
            Lua.DoString("LFGListInviteDialog.AcceptButton:Click(); LFGListInviteDialog.AcknowledgeButton:Click();");
        }

        private static CategoryInfo GetCategoryInfo(int categoryId)
        {
            var a = Lua.GetReturnValues($"return C_LFGList.GetCategoryInfo({categoryId});");

            var name = Lua.ParseLuaValue<string>(a[0]);
            var seperate = Lua.ParseLuaValue<bool>(a[1]);
            var auto = Lua.ParseLuaValue<bool>(a[2]);
            var prefer = Lua.ParseLuaValue<bool>(a[3]);

            return new CategoryInfo
            {
                Category = (LFGCategory) categoryId,
                Name = name,
                SeparateRecommended = seperate,
                AutoChoose = auto,
                PreferCurrentArea = prefer
            };
        }

        /// <summary>
        ///     Initializes a search for groups within the category
        /// </summary>
        /// <param name="categorieId">Category Id</param>
        public static void Search(int categorieId)
        {
            Lua.DoString(
                "C_LFGList.ClearSearchResults(); " +
                "local languages = C_LFGList.GetLanguageSearchFilter(); " +
                $"C_LFGList.Search({categorieId}, '', 0, 0, languages);");
        }

        /// <summary>
        ///     Initializes a search for groups within the category
        /// </summary>
        /// <param name="category">CategoryInfo</param>
        public static void Search(CategoryInfo category)
        {
            Log($"Starting a search of {category}");
            Search(category.Category);
        }

        /// <summary>
        ///     Starts the search for groups
        /// </summary>
        /// <param name="category">The category you wish to search</param>
        public static void Search(LFGCategory category)
        {
            Log($"Starting a search of {category}");
            Search((int) category);
        }

        private static ApplicationInfo GetApplicationInfo(int applicationId)
        {
            var info = Lua.GetReturnValues($"return C_LFGList.GetApplicationInfo({applicationId});");

            return new ApplicationInfo
            {
                Id = Lua.ParseLuaValue<int>(info[0]),
                Status = GetApplicationState(Lua.ParseLuaValue<string>(info[1])),
                PendingStatus = Lua.ParseLuaValue<bool>(info[2]),
                ApplicationDuration = TimeSpan.FromMinutes(Lua.ParseLuaValue<double>(info[3])),
                AccpetedRole = Lua.ParseLuaValue<string>(info[4])
            };
        }

        private static SearchResultInfo GetSearchResultInfo(int resultId)
        {
            var r = Lua.GetReturnValues($"return C_LFGList.GetSearchResultInfo({resultId});");
            //retrieve the realm & player name to see if they check out
            var s = Lua.ParseLuaValue<string>(r[11]);

            if (s == null)
                return new SearchResultInfo
                {
                    Id = resultId,
                    ActivityId = Lua.ParseLuaValue<int>(r[1]),
                    Name = Lua.ParseLuaValue<string>(r[2]),
                    Comment = Lua.ParseLuaValue<string>(r[3]),
                    VoiceChat = Lua.ParseLuaValue<bool>(r[4]),
                    iLevel = Lua.ParseLuaValue<int>(r[5]),
                    Age = TimeSpan.FromSeconds(Lua.ParseLuaValue<int>(r[6])),
                    NumBNetFriends = Lua.ParseLuaValue<int>(r[7]),
                    NumCharFriends = Lua.ParseLuaValue<int>(r[8]),
                    NumGuildMates = Lua.ParseLuaValue<int>(r[9]),
                    IsDelisted = Lua.ParseLuaValue<bool>(r[10]),
                    PlayerCount = Lua.ParseLuaValue<int>(r[12]),
                    AutoInvite = Lua.ParseLuaValue<bool>(r[13]),
                    Realm = "null",
                    LeaderName = "null"
                };

            char[] sep = {'-'};
            var wowSplit = s.Split(sep, 2);

            var l = wowSplit.Length < 1 ? "null" : wowSplit[0];
            var ra = wowSplit.Length < 2 ? "null" : wowSplit[1];

            return new SearchResultInfo
            {
                Id = resultId,
                ActivityId = Lua.ParseLuaValue<int>(r[1]),
                Name = Lua.ParseLuaValue<string>(r[2]),
                Comment = Lua.ParseLuaValue<string>(r[3]),
                VoiceChat = Lua.ParseLuaValue<bool>(r[4]),
                iLevel = Lua.ParseLuaValue<int>(r[5]),
                Age = TimeSpan.FromSeconds(Lua.ParseLuaValue<int>(r[6])),
                NumBNetFriends = Lua.ParseLuaValue<int>(r[7]),
                NumCharFriends = Lua.ParseLuaValue<int>(r[8]),
                NumGuildMates = Lua.ParseLuaValue<int>(r[9]),
                IsDelisted = Lua.ParseLuaValue<bool>(r[10]),
                PlayerCount = Lua.ParseLuaValue<int>(r[12]),
                AutoInvite = Lua.ParseLuaValue<bool>(r[13]),
                LeaderName = l,
                Realm = ra
            };
        }

        public static bool LeaveGroup()
        {
            Log("Leaving Current Group");
            Lua.DoString("LeaveParty();");
            var s = new SleepForLagDuration();
            s.ExecuteCoroutine();
            return StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid;
        }

        /// <summary>
        ///     Casts application state to enum ApplicationState
        /// </summary>
        /// <param name="application"></param>
        /// <returns></returns>
        private static ApplicationState GetApplicationState(string application)
        {
            switch (application)
            {
                case "invited":
                    return ApplicationState.Invited;
                case "failed":
                    return ApplicationState.Failed;
                case "cancelled":
                    return ApplicationState.Cancelled;
                case "applied":
                    return ApplicationState.Applied;
                case "inviteaccepted":
                    return ApplicationState.InviteAccepted;
                case "none":
                    return ApplicationState.None;

                default:
                {
                    Log($"{application} not listed as a state");
                    return ApplicationState.Failed;
                }
            }
        }

        private static readonly Dictionary<int, ActivityInfo> CachedActivityInfos = new Dictionary<int, ActivityInfo>(); 

        /// <summary>
        ///     Get the activity info by id
        /// </summary>
        /// <param name="activityId">Activity Id</param>
        /// <returns>Active info for the activity id</returns>
        public static ActivityInfo GetActivityInfo(int activityId)
        {
            if (CachedActivityInfos.ContainsKey(activityId)) return CachedActivityInfos[activityId];

            // //dump C_LFGList.GetActivityInfo(3)
            var info = Lua.GetReturnValues($"return C_LFGList.GetActivityInfo({activityId});");

            var r = new ActivityInfo
            {
                //fullName, shortName, categoryID, groupID, itemLevel, filters, minLevel, maxPlayers, displayType, activityOrder
                FullName = info[0],
                ShortName = info[1],
                Category = (LFGCategory) Lua.ParseLuaValue<int>(info[2]),
                GroupId = Lua.ParseLuaValue<int>(info[3]),
                ItemLevel = Lua.ParseLuaValue<int>(info[4]),
                MinLevel = Lua.ParseLuaValue<int>(info[6]),
                MaxPlayers = Lua.ParseLuaValue<int>(info[7]),
                ActivityId = activityId
            };

            CachedActivityInfos.Add(activityId, r);
            return r;
        }
    }

    [SuppressMessage("ReSharper", "InconsistentNaming")]
    public enum LFGCategory
    {
        None = 0,
        Questing = 1,
        Dungeons = 2,
        Raids = 3,
        Arenas = 4,
        Scenarios = 5,
        Custom = 6,
        Skirmishes = 7,
        Battlegrounds = 8,
        RatedBGs = 9,
        Ashran = 10
    }

    public class ActivityInfo
    {
        public LFGCategory Category { get; set; }
        public int ActivityId { get; set; }
        public string FullName { get; set; }
        public int GroupId { get; set; }
        public int ItemLevel { get; set; }
        public int MaxPlayers { get; set; }
        public int MinLevel { get; set; }
        public string ShortName { get; set; }

        public override string ToString()
        {
            return $"ActivityInfo {FullName} (Id {ActivityId})";
        }

        public override bool Equals(object obj)
        {
            return base.Equals(obj);
        }

        protected bool Equals(ActivityInfo other)
        {
            return ActivityId == other.ActivityId;
        }

        public override int GetHashCode()
        {
            return ActivityId;
        }
    }

    [Flags]
    public enum ApplicationState
    {
        Invited,
        Failed,
        Cancelled,
        Applied,
        InviteAccepted,
        None
    }

    public class CategoryInfo 
    {
        public bool AutoChoose { get; set; }
        public string Name { get; set; }
        public bool PreferCurrentArea { get; set; }
        public bool SeparateRecommended { get; set; }

        public CategoryInfo()
        {
            LastSearchTime = DateTime.MinValue;
        }

        public LFGCategory Category { get; set; }

        public DateTime LastSearchTime { get; set; }

        /// <summary>
        ///     Default entry for category info
        /// </summary>
        public static CategoryInfo None => new CategoryInfo
        {
            AutoChoose = false,
            Category = LFGCategory.None,
            LastSearchTime = DateTime.MinValue,
            Name = "None",
            PreferCurrentArea = false,
            SeparateRecommended = false
        };

        public void SearchCatergory()
        {
            LfgList.Search(this);
            LastSearchTime = DateTime.Now;
        }

        public override string ToString()
        {
            return $"{Name}";
        }

        public override bool Equals(object obj)
        {
            return base.Equals(obj);
        }

        protected bool Equals(CategoryInfo other)
        {
            return Category == other.Category;
        }

        public override int GetHashCode()
        {
            return (int) Category;
        }
    }

    public class ApplicationInfo
    {
        public string AccpetedRole { get; set; }
        public TimeSpan ApplicationDuration { get; set; }
        public int Id { get; set; }
        public bool PendingStatus { get; set; }
        public ApplicationState Status { get; set; }

        public override string ToString()
        {
            return $"ApplicationId {Id} {Status}";
        }

        public override bool Equals(object obj)
        {
            return base.Equals(obj);
        }

        protected bool Equals(ApplicationInfo other)
        {
            return Id == other.Id;
        }

        protected bool Equals(SearchResultInfo other)
        {
            return Id == other.Id;
        }

        public override int GetHashCode()
        {
            return Id;
        }
    }

    [SuppressMessage("ReSharper", "InconsistentNaming")]
    public class SearchResultInfo
    {
        public int ActivityId { get; set; }

        private ActivityInfo _cActivityInfo;
        public ActivityInfo ActivityInfo => _cActivityInfo ?? (_cActivityInfo = LfgList.GetActivityInfo(ActivityId));

        public TimeSpan Age { get; set; }
        public bool AutoInvite { get; set; }
        public string Comment { get; set; }
        public int Id { get; set; }
        public int iLevel { get; set; }
        public bool IsDelisted { get; set; }
        public string LeaderName { get; set; }
        public string Name { get; set; }
        public int NumBNetFriends { get; set; }
        public int NumCharFriends { get; set; }
        public int NumGuildMates { get; set; }
        public int PlayerCount { get; set; }
        public string Realm { get; set; }
        public bool VoiceChat { get; set; }

        public void ApplyToGroup()
        {
            LfgList.ApplyToGroup(this);
        }

        public override string ToString()
        {
            return $"{Name} on {Realm} ({Id}, {ActivityInfo.FullName}, AutoInvite: {AutoInvite})";
        }

        public override bool Equals(object obj)
        {
            return base.Equals(obj);
        }

        protected bool Equals(SearchResultInfo other)
        {
            return Id == other.Id;
        }

        protected bool Equals(ApplicationInfo other)
        {
            return Id == other.Id;
        }

        public override int GetHashCode()
        {
            return Id;
        }
    }
}