using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Windows.Media;
using Styx;
using Styx.Common;
using Styx.Plugins;
using Styx.WoWInternals;

namespace NuokRealmHop
{
    internal class Plugin : HBPlugin
    {
        public static List<CategoryInfo> AvailableCategories = new List<CategoryInfo>();
        public static List<ApplicationInfo> ApplicationInfos = new List<ApplicationInfo>();
        private static List<SearchResultInfo> _searchResults = new List<SearchResultInfo>();

        private static CategoryInfo _lastSearchCategory = CategoryInfo.None;
        private static bool _isWaitingForResults;
        private static bool _resultsUpdated;

        private static readonly Stopwatch JumpTimer = new Stopwatch();
        private static readonly Dictionary<string, DateTime> ServerTracker = new Dictionary<string, DateTime>();

        private static readonly Stopwatch SearchTimer = new Stopwatch();
        private static int _pulseCounter;
        private static readonly HashSet<int> HasAppliedId = new HashSet<int>();

        public override string Name => "NuokRealmHop";
        public override string Author => "Nuok";
        public override Version Version => new Version(3, 8);
        public override bool WantButton => true;

        private static void Log(string format, params object[] args)
        {
            Logging.WriteDiagnostic(Colors.Azure, "NuokRealmHop: " + format, args);
        }

        public override void OnEnable()
        {
            Log($"{Name} {Version} Enabled");
            LfgList.LFGLiftFrame_Show();
            Lua.Events.AttachEvent("LFG_LIST_SEARCH_FAILED", SearchFailed);
            Lua.Events.AttachEvent("LFG_LIST_SEARCH_RESULTS_RECEIVED", ResultsRecived);
            Lua.Events.AttachEvent("LFG_LIST_SEARCH_RESULT_UPDATED", ResultsUpdated);
            AvailableCategories = LfgList.GetAvailableCategories;
            _isWaitingForResults = false;
            _resultsUpdated = false;
            _searchResults.Clear();
            ApplicationInfos.Clear();
            _lastSearchCategory = CategoryInfo.None;

            var aCat = AvailableCategories.Aggregate("", (current, c) => current + $"{c.Name} ");
            Log(aCat);
        }

        private void SearchFailed(object sender, LuaEventArgs args)
        {
            if (NSettings.Instance.DebugDetailed)
                Log($"LFG_LIST_SEARCH_FAILED {_lastSearchCategory}");
            //Get the current search cat and reset its last search time to min
            var info = AvailableCategories.FirstOrDefault(c => c == _lastSearchCategory);
            if (info != null)
                info.LastSearchTime = DateTime.MinValue;
            _lastSearchCategory = CategoryInfo.None;
            _isWaitingForResults = false;
            _resultsUpdated = false;
        }


        private void ResultsUpdated(object sender, LuaEventArgs args)
        {
            if (NSettings.Instance.DebugDetailed)
                Log("LFG_LIST_SEARCH_RESULT_UPDATED");
            _resultsUpdated = true;
            _isWaitingForResults = false;
        }

        public override void OnDisable()
        {
            Log($"{Name} {Version} Disabled");
            Lua.Events.DetachEvent("LFG_LIST_SEARCH_FAILED", SearchFailed);
            Lua.Events.DetachEvent("LFG_LIST_SEARCH_RESULTS_RECEIVED", ResultsRecived);
            Lua.Events.DetachEvent("LFG_LIST_SEARCH_RESULT_UPDATED", ResultsUpdated);
            _searchResults.Clear();
            _pulseCounter = 0;

            _isWaitingForResults = false;
            _resultsUpdated = false;
            _searchResults.Clear();
            ApplicationInfos.Clear();
            SearchTimer.Stop();
            JumpTimer.Stop();
        }

        public override void OnButtonPress()
        {
            string logTag = $"{Name} v{Version}";
            var window = new ConfigWindow(Name, logTag, Author, 350, 400);
            window.Show();
            window.Closed += ConfigWindowOnClosed;
        }

        private void ConfigWindowOnClosed(object sender, EventArgs e)
        {
            NSettings.Instance.Save();
            Log("Settings Saved");
        }

        private void ResultsRecived(object sender, LuaEventArgs args)
        {
            using (new PerformanceLogger(NSettings.Instance.DebugPerformace, "ResultsRecived"))
            {
                if (NSettings.Instance.DebugDetailed)
                    Log("LFG_LIST_SEARCH_RESULTS_RECEIVED");
                _searchResults = LfgList.GetSearchResults;
                var cSearch = AvailableCategories.FirstOrDefault(c => c == _lastSearchCategory);
                if (cSearch != null)
                    cSearch.LastSearchTime = DateTime.Now;
            }
        }

        private static bool CategoryEnabled(CategoryInfo category)
        {
            var s = NSettings.Instance;
            switch (category.Category)
            {
                case LFGCategory.None:
                    return false;
                case LFGCategory.Arenas:
                    return s.Arenas;
                case LFGCategory.Ashran:
                    return s.Ashran;
                case LFGCategory.Battlegrounds:
                    return s.Battlegrounds;
                case LFGCategory.Custom:
                    return s.Custom;
                case LFGCategory.Dungeons:
                    return s.Dungeons;
                case LFGCategory.Questing:
                    return s.Questing;
                case LFGCategory.Raids:
                    return s.Raids;
                case LFGCategory.Scenarios:
                    return s.Scenarios;
                case LFGCategory.Skirmishes:
                    return s.Skirmishes;
            }
            return false;
        }

        private static bool ShouldHopToRealm(string realm)
        {
            if (!ServerTracker.ContainsKey(realm))
                return true;
            var t = Math.Round(DateTime.Now.Subtract(ServerTracker[realm]).TotalMinutes, 2);
            var b = t > NSettings.Instance.RealmBlacklist;

            if (NSettings.Instance.DebugDetailed)
            {
                Log(b
                    ? $"Should Hop to Realm {realm} ({t}min since last on realm"
                    : $"Realm {realm} currently blacklisted, {t}min since last on realm");
            }

            return b;
        }

        public override void Pulse()
        {
            var s = NSettings.Instance;
            _pulseCounter++;
            if (_pulseCounter < s.DelayPulse) return;
            _pulseCounter = 0;

            using (new PerformanceLogger(s.DebugPerformace, "Pulse"))
            {
                if (AvailableCategories.Count < 1) AvailableCategories = LfgList.GetAvailableCategories;
                var cRealm = LfgList.GetRealmName;

                if (cRealm == null)
                {
                    if (s.DebugDetailed)
                        Log("GetRealmName returned null");
                    return;
                }
                if (ServerTracker.ContainsKey(cRealm))
                    ServerTracker[cRealm] = DateTime.Now;
                else
                    ServerTracker.Add(cRealm, DateTime.Now);

                //if were in a party lets check if we should leave after a certain time or continue on
                if (StyxWoW.Me.GroupInfo.IsInParty || StyxWoW.Me.GroupInfo.IsInRaid)
                {
                    if (StyxWoW.Me.GroupInfo.GroupLeader == StyxWoW.Me && s.LeaderLeaver)
                    {
                        Log("Leaving the group as your its leader");
                        LfgList.LeaveGroup();
                        return;
                    }
                    if (s.LeaveAfterMin <= 0 && s.LeaveAfterSeconds <= 0) return;
                    if (!JumpTimer.IsRunning)
                    {
                        JumpTimer.Restart();
                        return;
                    }

                    var time = new TimeSpan(0, s.LeaveAfterMin, s.LeaveAfterSeconds);
                    var jumpIn = time.Subtract(JumpTimer.Elapsed);

                    if (s.DebugDetailed)
                        Log($"Time till next jump {jumpIn.TotalMinutes} minutes {jumpIn.TotalSeconds} seconds");
                    if (jumpIn.TotalSeconds > 0) return;

                    Log("Leaving group to Jump Realm");
                    if (!LfgList.LeaveGroup()) return;
                    JumpTimer.Stop();
                    ApplicationInfos = new List<ApplicationInfo>();
                    _searchResults.Clear();
                    return;
                }

                if (_resultsUpdated)
                {
                    ApplicationInfos = LfgList.GetApplicationInfos;
                    if (s.DebugDetailed)
                        Log($"LFG_LIST_SEARCH_RESULT_UPDATED found {ApplicationInfos.Count} applications");
                    if (ApplicationInfos.Any(a => a.Status == ApplicationState.Invited))
                    {
                        LfgList.AcceptPopUps();
                    }
                    _resultsUpdated = false;
                }

                var applications = LfgList.NumActiveAppliications;

                if (s.DebugDetailed)
                    Log($"NumActiveApplications: {applications}");

                if (applications >= 5) return;

                var now = DateTime.Now;

                foreach (var r in from r in _searchResults
                    where !HasAppliedId.Contains(r.Id)
                    where !r.IsDelisted
                    where !ApplicationInfos.Exists(a => a.Id == r.Id)
                    where !(r.iLevel > StyxWoW.Me.AverageItemLevelEquipped)
                    where !s.AutoInvite || r.AutoInvite
                    where r.PlayerCount <= s.MaxPartySize
                    where ShouldHopToRealm(r.Realm)
                    select r)
                {
                    HasAppliedId.Add(r.Id);
                    r.ApplyToGroup();
                    applications = LfgList.NumActiveAppliications;
                    if (applications >= 5) break;
                }

                //If were searching wait until it finishes or times out
                if (_isWaitingForResults && _lastSearchCategory != null &&
                    _lastSearchCategory.Category != LFGCategory.None)
                {
                    if (!SearchTimer.IsRunning) SearchTimer.Restart();
                    if (!(SearchTimer.Elapsed.TotalSeconds > 5)) return;
                    Log($"{_lastSearchCategory} search has timed out, researching");

                    //Get the current search cat and reset its last search time to min
                    var info = AvailableCategories.FirstOrDefault(c => c == _lastSearchCategory);
                    if (info != null)
                        info.LastSearchTime = DateTime.MinValue;

                    _lastSearchCategory = CategoryInfo.None;
                    SearchTimer.Stop();
                    return;
                }
                SearchTimer.Stop();

                if (s.DebugDetailed)
                    foreach (var cat in AvailableCategories)
                    {
                        var dif = now.Subtract(cat.LastSearchTime).TotalSeconds;
                        Log($@"{cat} last search time is {cat.LastSearchTime}, {dif}");
                    }

                //Start a new search if not searching
                foreach (var cat in AvailableCategories.Where(
                    cat => CategoryEnabled(cat) &&
                           now.Subtract(cat.LastSearchTime).TotalSeconds > s.CatSearchDelay))
                {
                    cat.SearchCatergory();
                    //Old results are not valid anymore anyway
                    _searchResults.Clear();
                    _lastSearchCategory = cat;
                    _isWaitingForResults = true;
                    break;
                }
            }
        }
    }
}